// 
// TiglUsb driver for Windows 9x/Me & 2000/XP
// Library (user mode) for talking with the driver (kernel mode)
//
// Copyright (c) 2001-2002 Romain Livin
// roms@lpg.ticalc.org
// http://lpg.ticalc.org/prj_usb
//
// July the 4th, 2002
//

/*++

Copyright (c) 2001-2002, Romain Livin. All rights reserved.

Module Name:

    Device.c (this file is the same for both drivers (98DDK & XPDDK))

Abstract:

    Library (user mode) used for talking with the driver (kernel mode).
	This unit contains a set of functions for:
	- obtaining the device name
	- opening an device instance
	- doing some IOCTL codes

Environment:

    user mode only

Notes:

  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
  KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
  PURPOSE.

  Copyright (c) 2001 Romain Livin.  All Rights Reserved.


Revision History:

	11/17/97: created
	04/10/01: modified
	28/12/01: modified
	02/04/02: removed & added some IOCTL access functions

--*/

#include <windows.h>

#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <time.h>

#include "devioctl.h"

#include <setupapi.h>
#include <basetyps.h>
#include "BulkUsr.h"

#include "usbdi.h"

#include "../api/tiglusb.h"


#define NOISY(_x_) printf _x_ ;

char completeDeviceName[256] = "";  //generated from the GUID registered by the driver itself 

/*++
Routine Description:

    Given the HardwareDeviceInfo, representing a handle to the plug and
    play information, and deviceInfoData, representing a specific usb device,
    open that device and fill in all the relevant information in the given
    USB_DEVICE_DESCRIPTOR structure.

Arguments:

    HardwareDeviceInfo:  handle to info obtained from Pnp mgr via SetupDiGetClassDevs()
    DeviceInfoData:      ptr to info obtained via SetupDiEnumInterfaceDevice()

Return Value:

    return HANDLE if the open and initialization was successfull,
	else INVLAID_HANDLE_VALUE.

--*/
HANDLE
OpenOneDevice (
    IN       HDEVINFO                    HardwareDeviceInfo,
    IN       PSP_INTERFACE_DEVICE_DATA   DeviceInfoData,
	IN		 char *devName
    )
{
    PSP_INTERFACE_DEVICE_DETAIL_DATA     functionClassDeviceData = NULL;
    ULONG                                predictedLength = 0;
    ULONG                                requiredLength = 0;
	HANDLE								 hOut = INVALID_HANDLE_VALUE;

    //
    // allocate a function class device data structure to receive the
    // goods about this particular device.
    //
    SetupDiGetInterfaceDeviceDetail (
            HardwareDeviceInfo,
            DeviceInfoData,
            NULL, // probing so no output buffer yet
            0, // probing so output buffer length of zero
            &requiredLength,
            NULL); // not interested in the specific dev-node


    predictedLength = requiredLength;
    // sizeof (SP_FNCLASS_DEVICE_DATA) + 512;

    functionClassDeviceData = malloc (predictedLength);
    functionClassDeviceData->cbSize = sizeof (SP_INTERFACE_DEVICE_DETAIL_DATA);

    //
    // Retrieve the information from Plug and Play.
    //
    if (! SetupDiGetInterfaceDeviceDetail (
               HardwareDeviceInfo,
               DeviceInfoData,
               functionClassDeviceData,
               predictedLength,
               &requiredLength,
               NULL)) {
		free( functionClassDeviceData );
        return INVALID_HANDLE_VALUE;
    }

	strcpy( devName,functionClassDeviceData->DevicePath) ;
	fprintf(stdout,  "TiglUsb.dll: Attempting to open %s\n", devName );

    hOut = CreateFile (
                  functionClassDeviceData->DevicePath,
                  GENERIC_READ | GENERIC_WRITE,
                  FILE_SHARE_READ | FILE_SHARE_WRITE,
                  NULL, // no SECURITY_ATTRIBUTES structure
                  OPEN_EXISTING, // No special create flags
                  0, // No special attributes
                  NULL); // No template file

    if (INVALID_HANDLE_VALUE == hOut) {
		fprintf(stdout,  "TiglUsb.dll: FAILED to open %s\n", devName );
    }
	free( functionClassDeviceData );
	return hOut;
}


/*++
Routine Description:

   Do the required PnP things in order to find 
   the next available proper device in the system at this time.

Arguments:

    pGuid:      ptr to GUID registered by the driver itself
    outNameBuf: the generated name for this device

Return Value:

    return HANDLE if the open and initialization was successful,
	else INVLAID_HANDLE_VALUE.
--*/
HANDLE
OpenUsbDevice( LPGUID  pGuid, char *outNameBuf)
{
   ULONG NumberDevices;
   HANDLE hOut = INVALID_HANDLE_VALUE;
   HDEVINFO                 hardwareDeviceInfo;
   SP_INTERFACE_DEVICE_DATA deviceInfoData;
   ULONG                    i;
   BOOLEAN                  done;
   PUSB_DEVICE_DESCRIPTOR   usbDeviceInst;
   PUSB_DEVICE_DESCRIPTOR	*UsbDevices = &usbDeviceInst;

   *UsbDevices = NULL;
   NumberDevices = 0;

   //
   // Open a handle to the plug and play dev node.
   // SetupDiGetClassDevs() returns a device information set that contains info on all 
   // installed devices of a specified class.
   //
   hardwareDeviceInfo = SetupDiGetClassDevs (
                           pGuid,
                           NULL, // Define no enumerator (global)
                           NULL, // Define no
                           (DIGCF_PRESENT | // Only Devices present
                            DIGCF_INTERFACEDEVICE)); // Function class devices.

   //
   // Take a wild guess at the number of devices we have;
   // Be prepared to realloc and retry if there are more than we guessed
   //
   NumberDevices = 4;
   done = FALSE;
   deviceInfoData.cbSize = sizeof (SP_INTERFACE_DEVICE_DATA);

   i=0;
   while (!done) {
      NumberDevices *= 2;

      if (*UsbDevices) {
         *UsbDevices =
               realloc (*UsbDevices, (NumberDevices * sizeof (USB_DEVICE_DESCRIPTOR)));
      } else {
         *UsbDevices = calloc (NumberDevices, sizeof (USB_DEVICE_DESCRIPTOR));
      }

      if (NULL == *UsbDevices) {

         // SetupDiDestroyDeviceInfoList destroys a device information set 
         // and frees all associated memory.

         SetupDiDestroyDeviceInfoList (hardwareDeviceInfo);
         return INVALID_HANDLE_VALUE;
      }

      usbDeviceInst = *UsbDevices + i;

      for (; i < NumberDevices; i++) {

         // SetupDiEnumDeviceInterfaces() returns information about device interfaces 
         // exposed by one or more devices. Each call returns information about one interface;
         // the routine can be called repeatedly to get information about several interfaces
         // exposed by one or more devices.

         if (SetupDiEnumDeviceInterfaces (hardwareDeviceInfo,
                                         0, // We don't care about specific PDOs
										 pGuid,
                                         i,
                                         &deviceInfoData)) {

            hOut = OpenOneDevice (hardwareDeviceInfo, &deviceInfoData, outNameBuf);
			if ( hOut != INVALID_HANDLE_VALUE ) {
               done = TRUE;
               break;
			}
         } else {
            if (ERROR_NO_MORE_ITEMS == GetLastError()) {
               done = TRUE;
               break;
            }
         }
      }
   }

   NumberDevices = i;

   // SetupDiDestroyDeviceInfoList() destroys a device information set 
   // and frees all associated memory.

   SetupDiDestroyDeviceInfoList (hardwareDeviceInfo);
   free ( *UsbDevices );
   return hOut;
}


/*++
Routine Description:

    Given a ptr to a driver-registered GUID, give us a string with the device name
    that can be used in a CreateFile() call.
    Actually briefly opens and closes the device and sets outBuf if successfull;
    returns FALSE if not 

Arguments:

    pGuid:      ptr to GUID registered by the driver itself
    outNameBuf: the generated zero-terminated name for this device

Return Value:

    TRUE on success else FALSE

--*/
BOOL
GetUsbDeviceFileName( LPGUID  pGuid, char *outNameBuf)
{
	HANDLE hDev = OpenUsbDevice( pGuid, outNameBuf );
	if ( hDev != INVALID_HANDLE_VALUE )
	{
		CloseHandle( hDev );
		return TRUE;
	}
	return FALSE;

}


/*++
Routine Description:

    Called by dumpUsbConfig() to open an instance of our device

Arguments:

    None

Return Value:

    Device handle on success else NULL

--*/
HANDLE
open_dev()
{

	HANDLE hDEV = OpenUsbDevice( (LPGUID)&GUID_CLASS_I82930_BULK, completeDeviceName);


	if (hDEV == INVALID_HANDLE_VALUE) {
		fprintf(stdout, "TiglUsb.dll: Failed to open (%s) = %d", completeDeviceName, GetLastError());
	} else {
		fprintf(stdout, "TiglUsb.dll: DeviceName = (%s)\n", completeDeviceName);
    }		

	return hDEV;
}


/*++
Routine Description:

    Called by main() to open an instance of our device after obtaining its name

Arguments:

    None

Return Value:

    Device handle on success else NULL

--*/
HANDLE
open_file( char *filename)
{

	int success = 1;
	HANDLE h;

	if ( !GetUsbDeviceFileName( 
		(LPGUID) &GUID_CLASS_I82930_BULK, 
		completeDeviceName) )
	{
		NOISY(("Failed to GetUsbDeviceFileName\n", GetLastError()));
		return  INVALID_HANDLE_VALUE; 
	}

    strcat (completeDeviceName,
			"\\"
			);			

    strcat (completeDeviceName,
			filename
			);					

	fprintf(stdout, "TiglUsb.dll: completeDeviceName = (%s)\n", completeDeviceName);

#ifndef NONBLOCKING
	fprintf(stdout, "TiglUsb.dll: Device open in non overlapped mode.\n");
	h = CreateFile(completeDeviceName,
		GENERIC_WRITE | GENERIC_READ,
		FILE_SHARE_WRITE | FILE_SHARE_READ,
		NULL,
		OPEN_EXISTING,
		0,
		NULL);
#else
	fprintf(stdout, "TiglUsb.dll: Device open in overlapped mode.\n");
	h = CreateFile(completeDeviceName,
		GENERIC_WRITE | GENERIC_READ,
		FILE_SHARE_WRITE | FILE_SHARE_READ,
		NULL,
		OPEN_EXISTING,
		FILE_FLAG_OVERLAPPED,
		NULL);
#endif

	if (h == INVALID_HANDLE_VALUE) {
		NOISY(("Failed to open (%s) = %d", completeDeviceName, GetLastError()));
		success = 0;
	} else {
			NOISY(("Opened successfully.\n"));
    }		

	return h;
}


// Begin, routines for USB configuration dump (Cmdline "rwbulk -u" )

/*++
Routine Description:

    Called to get ascii string of USB descriptor 

Arguments:

	PUSB_ENDPOINT_DESCRIPTOR->bDescriptorType or
	PUSB_DEVICE_DESCRIPTOR->bDescriptorType or
	PUSB_INTERFACE_DESCRIPTOR->bDescriptorType or
	PUSB_STRING_DESCRIPTOR->bDescriptorType or
	PUSB_POWER_DESCRIPTOR->bDescriptorType or
	PUSB_CONFIGURATION_DESCRIPTOR->bDescriptorType 

Return Value:

    ptr to string

--*/
char
*usbDescriptorTypeString(UCHAR bDescriptorType )
{

	switch(bDescriptorType) {

	case USB_DEVICE_DESCRIPTOR_TYPE:
		return "USB_DEVICE_DESCRIPTOR_TYPE";

	case USB_CONFIGURATION_DESCRIPTOR_TYPE:
		return "USB_CONFIGURATION_DESCRIPTOR_TYPE";
		

	case USB_STRING_DESCRIPTOR_TYPE:
		return "USB_STRING_DESCRIPTOR_TYPE";
		

	case USB_INTERFACE_DESCRIPTOR_TYPE:
		return "USB_INTERFACE_DESCRIPTOR_TYPE";
		

	case USB_ENDPOINT_DESCRIPTOR_TYPE:
		return "USB_ENDPOINT_DESCRIPTOR_TYPE";
		

#ifdef USB_POWER_DESCRIPTOR_TYPE // this is the older definintion which is actually obsolete
    // workaround for temporary bug in 98ddk, older USB100.h file
	case USB_POWER_DESCRIPTOR_TYPE:
		return "USB_POWER_DESCRIPTOR_TYPE";
#endif
		
#ifdef USB_RESERVED_DESCRIPTOR_TYPE  // this is the current version of USB100.h as in NT5DDK

	case USB_RESERVED_DESCRIPTOR_TYPE:
		return "USB_RESERVED_DESCRIPTOR_TYPE"; 
             
	case USB_CONFIG_POWER_DESCRIPTOR_TYPE:
		return "USB_CONFIG_POWER_DESCRIPTOR_TYPE";          
          
	case USB_INTERFACE_POWER_DESCRIPTOR_TYPE:       
		return "USB_INTERFACE_POWER_DESCRIPTOR_TYPE";       
#endif // for current nt5ddk version of USB100.h
        
	default:
		return "??? UNKNOWN!!";	
	}
}


/*++
Routine Description:

    Called to get ascii string of endpt descriptor type

Arguments:

	PUSB_ENDPOINT_DESCRIPTOR->bmAttributes

Return Value:

    ptr to string

--*/
char
*usbEndPointTypeString(UCHAR bmAttributes)
{
	UINT typ = bmAttributes & USB_ENDPOINT_TYPE_MASK;


	switch( typ) {
	case USB_ENDPOINT_TYPE_INTERRUPT:
		return "USB_ENDPOINT_TYPE_INTERRUPT";

	case USB_ENDPOINT_TYPE_BULK:
		return "USB_ENDPOINT_TYPE_BULK";	

	case USB_ENDPOINT_TYPE_ISOCHRONOUS:
		return "USB_ENDPOINT_TYPE_ISOCHRONOUS";	
		
	case USB_ENDPOINT_TYPE_CONTROL:
		return "USB_ENDPOINT_TYPE_CONTROL";	
		
	default:
		return "??? UNKNOWN!!";	
	}
}


/*++
Routine Description:

    Called to get ascii string of USB_CONFIGURATION_DESCRIPTOR attributes

Arguments:

	PUSB_CONFIGURATION_DESCRIPTOR->bmAttributes

Return Value:

    ptr to string

--*/
char
*usbConfigAttributesString(UCHAR bmAttributes)
{
	UINT typ = bmAttributes & USB_CONFIG_POWERED_MASK;


	switch( typ) {

	case USB_CONFIG_BUS_POWERED:
		return "USB_CONFIG_BUS_POWERED";

	case USB_CONFIG_SELF_POWERED:
		return "USB_CONFIG_SELF_POWERED";
		
	case USB_CONFIG_REMOTE_WAKEUP:
		return "USB_CONFIG_REMOTE_WAKEUP";

		
	default:
		return "??? UNKNOWN!!";	
	}
}


/*++
Routine Description:

    Called to do formatted ascii dump to console of a USB config descriptor 

Arguments:

    ptr to USB configuration descriptor

Return Value:

    none

--*/
void 
print_USB_CONFIGURATION_DESCRIPTOR(PUSB_CONFIGURATION_DESCRIPTOR cd)
{
    fprintf(stdout, "\n===================\nUSB_CONFIGURATION_DESCRIPTOR\n");

    fprintf(stdout, 
    "bLength = 0x%x, decimal %d\n", cd->bLength, cd->bLength
    );

    fprintf(stdout,     
    "bDescriptorType = 0x%x ( %s )\n", cd->bDescriptorType, usbDescriptorTypeString( cd->bDescriptorType ) 
    );

    fprintf(stdout,         
    "wTotalLength = 0x%x, decimal %d\n", cd->wTotalLength, cd->wTotalLength
    );

    fprintf(stdout,     
    "bNumInterfaces = 0x%x, decimal %d\n", cd->bNumInterfaces, cd->bNumInterfaces
    );

    fprintf(stdout,     
    "bConfigurationValue = 0x%x, decimal %d\n", cd->bConfigurationValue, cd->bConfigurationValue
    );

    fprintf(stdout,     
    "iConfiguration = 0x%x, decimal %d\n", cd->iConfiguration, cd->iConfiguration
    );

    fprintf(stdout,     
    "bmAttributes = 0x%x ( %s )\n", cd->bmAttributes, usbConfigAttributesString( cd->bmAttributes )
    );

    fprintf(stdout,     
    "MaxPower = 0x%x, decimal %d\n", cd->MaxPower, cd->MaxPower    
    );
}


/*++
Routine Description:

    Called to do formatted ascii dump to console of a USB interface descriptor 

Arguments:

    ptr to USB interface descriptor

Return Value:

    none

--*/
void 
print_USB_INTERFACE_DESCRIPTOR(PUSB_INTERFACE_DESCRIPTOR id, UINT ix)
{
    fprintf(stdout, "\n-----------------------------\nUSB_INTERFACE_DESCRIPTOR #%d\n", ix);


    fprintf(stdout, 
    "bLength = 0x%x\n", id->bLength
    );


    fprintf(stdout,     
    "bDescriptorType = 0x%x ( %s )\n", id->bDescriptorType, usbDescriptorTypeString( id->bDescriptorType ) 
    );


    fprintf(stdout,         
    "bInterfaceNumber = 0x%x\n", id->bInterfaceNumber
    );
    fprintf(stdout,     
    "bAlternateSetting = 0x%x\n", id->bAlternateSetting
    );
    fprintf(stdout,     
    "bNumEndpoints = 0x%x\n", id->bNumEndpoints
    );
    fprintf(stdout,     
    "bInterfaceClass = 0x%x\n", id->bInterfaceClass
    );
    fprintf(stdout,     
    "bInterfaceSubClass = 0x%x\n", id->bInterfaceSubClass
    );
    fprintf(stdout,     
    "bInterfaceProtocol = 0x%x\n", id->bInterfaceProtocol    
    );
    fprintf(stdout,     
    "bInterface = 0x%x\n", id->iInterface    
    );
}


/*++
Routine Description:

    Called to do formatted ascii dump to console of a USB endpoint descriptor 

Arguments:

    ptr to USB endpoint descriptor, 
	index of this endpt in interface desc

Return Value:

    none

--*/
void 
print_USB_ENDPOINT_DESCRIPTOR(PUSB_ENDPOINT_DESCRIPTOR ed, int i)
{
    fprintf(stdout, 
	"------------------------------\nUSB_ENDPOINT_DESCRIPTOR for Pipe%02d\n", i
	);

    fprintf(stdout, 
    "bLength = 0x%x\n", ed->bLength
    );

    fprintf(stdout,     
    "bDescriptorType = 0x%x ( %s )\n", ed->bDescriptorType, usbDescriptorTypeString( ed->bDescriptorType ) 
    );


	if ( USB_ENDPOINT_DIRECTION_IN( ed->bEndpointAddress ) ) {
		fprintf(stdout,         
		"bEndpointAddress= 0x%x ( INPUT )\n", ed->bEndpointAddress
		);
	} else {
		fprintf(stdout,         
		"bEndpointAddress= 0x%x ( OUTPUT )\n", ed->bEndpointAddress
		);
	}

    fprintf(stdout,     
    "bmAttributes= 0x%x ( %s )\n", ed->bmAttributes, usbEndPointTypeString ( ed->bmAttributes ) 
    );


    fprintf(stdout,     
    "wMaxPacketSize= 0x%x, decimal %d\n", ed->wMaxPacketSize, ed->wMaxPacketSize
    );
    fprintf(stdout,     
    "bInterval = 0x%x, decimal %d\n", ed->bInterval, ed->bInterval
    );
}


/*++
Routine Description:

    Called to do formatted ascii dump to console of  USB 
    configuration, interface, and endpoint descriptors 
    (Cmdline "rwbulk -u" )

Arguments:

    handle to device

Return Value:

    none

--*/
void
rw_dev( HANDLE hDEV )
{
	UINT success;
	int siz, nBytes;
	char buf[256];
    PUSB_CONFIGURATION_DESCRIPTOR cd;
    PUSB_INTERFACE_DESCRIPTOR id;
    PUSB_ENDPOINT_DESCRIPTOR ed;

	siz = sizeof(buf);

	if (hDEV == INVALID_HANDLE_VALUE) {
		NOISY(("DEV not open"));
		return;
	}
	
	success = DeviceIoControl(hDEV,
			IOCTL_BULKUSB_GET_CONFIG_DESCRIPTOR,
			buf,
			siz,
			buf,
			siz,
			&nBytes,
			NULL);

	NOISY(("IOCTL_TIGLUSB_GET_CONFIG_DESCRIPTOR request complete, success = %d nBytes = %d\n", success, nBytes));
	
	if (success) {
        ULONG i;
		UINT  j, n;
        char *pch;
        
        pch = buf;
		n = 0;

        cd = (PUSB_CONFIGURATION_DESCRIPTOR) pch;

        print_USB_CONFIGURATION_DESCRIPTOR( cd );

        pch += cd->bLength;

        do {
            
            id = (PUSB_INTERFACE_DESCRIPTOR) pch;

            print_USB_INTERFACE_DESCRIPTOR(id, n++);

            pch += id->bLength;
            for (j=0; j<id->bNumEndpoints; j++) {

                ed = (PUSB_ENDPOINT_DESCRIPTOR) pch; 

                print_USB_ENDPOINT_DESCRIPTOR(ed,j);

                pch += ed->bLength;
            }
            i = pch - buf;
        } while (i<cd->wTotalLength);       
        
	}
	
	return;

}


/*++
Routine Description:

    Called to do formatted ascii dump to console of  USB 
    configuration, interface, and endpoint descriptors 
    (Cmdline "rwbulk -u" )

Arguments:

    none

Return Value:

    none

--*/
int  dumpUsbConfig()
{

	HANDLE hDEV = open_dev();

	if ( hDEV )
	{
		rw_dev( hDEV );
		CloseHandle(hDEV);
	}

	return 0;
}
//  End, routines for USB configuration and pipe info dump  (Cmdline "rwbulk -u" ) 


/*++
Routine Description:

    Try to reset the pipe by using IOCTL codes.

Arguments:

    none

Return Value:

    none

--*/
int  resetPipe()
{
	UINT success;
	int siz, nBytes;
	char buf[256];

	HANDLE hDEV = open_dev();

	if ( hDEV )
	{
		siz = sizeof(buf);

		if (hDEV == INVALID_HANDLE_VALUE) {
			NOISY(("DEV not open"));
			return -1;
		}

		success = DeviceIoControl(hDEV,
			IOCTL_BULKUSB_RESET_PIPE,
			buf,
			siz,
			buf,
			siz,
			&nBytes,
			NULL);

		fprintf(stdout, "TiglUsb.dll: resetPipe\n");
		NOISY(("TiglUsb.dll: IOCTL_TIGLUSB_RESET_PIPE request complete, success = %s nBytes = %d\n", success ? "ok" : "nok", nBytes));
		CloseHandle(hDEV);
		if(success) return 0;
	}

	return -1;
}


/*++
Routine Description:

    Try to reset the device driver by using IOCTL codes.
	Beware: dangerous to use, system can become unstable (3, 2, 1, boum ?!)

Arguments:

    none

Return Value:

    none

--*/
int  resetDevice()
{
	UINT success;
	int siz, nBytes;
	char buf[256];

	HANDLE hDEV = open_dev();

	if ( hDEV )
	{
		siz = sizeof(buf);

		if (hDEV == INVALID_HANDLE_VALUE) {
			NOISY(("DEV not open"));
			return -1;
		}

		success = DeviceIoControl(hDEV,
			IOCTL_BULKUSB_RESET_DEVICE,
			buf,
			siz,
			buf,
			siz,
			&nBytes,
			NULL);

		NOISY(("IOCTL_TIGLUSB_RESET_DEVICE request complete, success = %s nBytes = %d\n", success ? "ok" : "nok", nBytes));
		CloseHandle(hDEV);
		if(success) return 0;
	}

	return -1;
}


/*++
Routine Description:

    Try to reset IN & OUT pipes by using IOCTL codes.

Arguments:

    none

Return Value:

    none

--*/
int  resetPipes()
{
	UINT success;
	int siz, nBytes;
	char buf[256];

	HANDLE hDEV = open_dev();

	if ( hDEV )
	{
		siz = sizeof(buf);

		if (hDEV == INVALID_HANDLE_VALUE) {
			NOISY(("DEV not open"));
			return -1;
		}

		success = DeviceIoControl(hDEV,
			IOCTL_BULKUSB_RESET_PIPES,
			buf,
			siz,
			buf,
			siz,
			&nBytes,
			NULL);

		fprintf(stdout, "TiglUsb.dll: resetPipes\n");
		NOISY(("TiglUsb.dll: IOCTL_TIGLUSB_RESET_PIPES request complete, success = %s nBytes = %d\n", success ? "ok" : "nok", nBytes));
		CloseHandle(hDEV);
		if(success) return 0;
	}

	return -1;
}

/*++
Routine Description:

    Copied from rw_dev. Called to get Product ID and Maximum Packet Size

Arguments:

    none

Return Value:

    none

--*/
int get_max_ps(int *max_ps)
{
	UINT success;
	int siz, nBytes;
	char buf[256];
	PUSB_CONFIGURATION_DESCRIPTOR cd;
	PUSB_INTERFACE_DESCRIPTOR id;
	PUSB_ENDPOINT_DESCRIPTOR ed;

	HANDLE hDEV = open_dev();

	*max_ps = 0;
	siz = sizeof(buf);

	if (hDEV == INVALID_HANDLE_VALUE) {
		NOISY(("DEV not open"));
		return -1;
	}
	
	success = DeviceIoControl(hDEV,
			IOCTL_BULKUSB_GET_CONFIG_DESCRIPTOR,
			buf,
			siz,
			buf,
			siz,
			&nBytes,
			NULL);

	NOISY(("IOCTL_TIGLUSB_GET_CONFIG_DESCRIPTOR request complete, success = %d nBytes = %d\n", success, nBytes));
	
	if (success) {

		UINT  j, n;
        char *pch;
        
        pch = buf;
		n = 0;

        cd = (PUSB_CONFIGURATION_DESCRIPTOR) pch;

        pch += cd->bLength;
           
        id = (PUSB_INTERFACE_DESCRIPTOR) pch;

        pch += id->bLength;
        for (j=0; j<id->bNumEndpoints; j++) {

            ed = (PUSB_ENDPOINT_DESCRIPTOR) pch; 

            pch += ed->bLength;
        }
        
	}

	*max_ps = ed->wMaxPacketSize;
	
	return 0;
}