// 
// TiglUsb driver for Windows 9x/Me, NT4/2000/XP
// Library (user mode) for talking with the driver (kernel mode)
//
// Copyright (c) 2001-2002 Romain Livin
// roms@lpg.ticalc.org
// http://lpg.ticalc.org/prj_usb
//
// July the 4th, 2002
//

/*++

Module Name:

    Device.c
	(this file is the same for both drivers (98DDK & XPDDK))

Abstract:

    Library (user mode) used for talking with the driver (kernel mode).
	This library contains a set of functions for:
	- opening/closing the driver
	- reading/writing with the driver
	- flushing buffers
	- setting/getting timeout
	- checking arrival of data (useful for emulators which need to monitor the linkport)

Environment:

    user mode only

Notes:

  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
  KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
  PURPOSE.

  Copyright (c) 2001-2005 Romain Livin.  All Rights Reserved.


Revision History:

	03/07/02: created
	23/08/02: overlapped I/O fixes
	06/10/02: added a function for polling data arrival & improvements
	21/12/02: added a function for retrieving library version
    11/04/04: improved to workaround the quirk (returns neither error nor data)
	21/04/04: improved to use heuristic-buffered writes (send is faster but still reliable)
	15/07/05: little clean-up and overhaul
--*/

/* 
   Some important remarks from <http://lpg.ticalc.org/prj_usb/SilverLink_Paper.txt>
   
   This link cable use Bulk mode with packets. The max size of a packet is 
   32 bytes (MAX_PACKET_SIZE/BULKUSB_MAX_TRANSFER_SIZE). 
   
   This is transparent for the user because the libusb manages all these 
   things for us. Nethertheless, this fact has some consequences:

   - it is better (for USB & OS performances) to read/write a set of bytes 
   rather than byte per byte.

   - for reading, we have to read up to 32 bytes at a time (even if we need 
   only 1 byte) and to store them in a buffer for subsequent accesses. 
   In fact, if we try and get byte per byte, it will not work.

   - for writing, we don't store bytes in a buffer. It seems better to send
   data byte per byte (latency ?!). But, this make data-rate significantly 
   decrease (1KB/s instead of 5KB/s).
   Another way is to use partially buffered write operations: send consecutive
   blocks as a whole but partial block byte per byte. This is the best 
   compromise and it works fine !

   - another particular effect (quirk): sometimes (usually when calc need to 
   reply and takes a while), a read call can returns with neither data 
   nor timeout. Simply retry a read call and it works fine. The best example 
   is to get IDLIST.
*/

#include <stdio.h>
#include <stdlib.h>

#include "stdafx.h"
#include "TiglUsb.h"

//
// DllMain : defines the entry point for the DLL application.
//
BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
    switch (ul_reason_for_call)
	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
    }
    return TRUE;
}

TIGLUSB_EXP char* TIGLUSB_API TiglUsbVersion(void)
{
	return "3.5b";
}

//
// TiglUsb.cpp: DLL routines
//

typedef DWORD                   TIME;
# define  toSTART(ref)          { (ref)=GetTickCount(); }
# define  toELAPSED(ref, max)   ( (int)(GetTickCount()-(ref)) > (100*max) )

static UINT max_ps = 0;

static HANDLE hDev = INVALID_HANDLE_VALUE;

static HANDLE hRead  = INVALID_HANDLE_VALUE;
static HANDLE hWrite = INVALID_HANDLE_VALUE;

static UCHAR *rBuf = NULL;
static DWORD nBytesRead = 0;

static UCHAR *wBuf = NULL;
static DWORD nBytesWrite = 0;

static int timeout = 15;	// default timeout: 1.5 seconds

static OVERLAPPED olp = { 0 };
static BOOL ioPending = FALSE;

static void print_last_error(void)
{
	LPVOID lpMsgBuf;
	
	FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
					FORMAT_MESSAGE_FROM_SYSTEM |
					FORMAT_MESSAGE_IGNORE_INSERTS,
					NULL, GetLastError(),
					MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
					(LPTSTR) &lpMsgBuf,    0,    NULL );
	fprintf(stdout, "TiglUsb.dll error: %i -> %s\n", GetLastError(), lpMsgBuf);
}

// Open the device (libtifiles: slv_init)
TIGLUSB_EXP int TIGLUSB_API TiglUsbOpen(void)
{
	if( (hRead != INVALID_HANDLE_VALUE) || (hWrite != INVALID_HANDLE_VALUE) )
		return -TIGLERR_DEV_ALREADY_OPEN;

	// Get maximum packet size and allocate buffers
	hDev = open_dev(1);
	if(hDev == INVALID_HANDLE_VALUE)
	{
		print_last_error();
		return -TIGLERR_DEV_OPEN_FAILED;
	}

	if(get_max_ps(hDev, &max_ps, NULL) == -1)
		return -TIGLERR_DEV_OPEN_FAILED;

	if(wBuf != NULL) free(wBuf);
	if(rBuf != NULL) free(rBuf);
	
	wBuf = (UCHAR *)malloc(max_ps * sizeof(UCHAR));
    rBuf = (UCHAR *)malloc(max_ps * sizeof(UCHAR));
	if((wBuf == NULL) || (rBuf == NULL))
    {
            free(wBuf); wBuf = NULL;
            free(rBuf); rBuf = NULL;
            return -TIGLERR_MALLOC;
    }
	fprintf(stdout,  "TiglUsb.dll: allocating %i bytes buffers\n", max_ps);

	// Open the USB device: 2 named pipes (endpoints)
	hWrite = open_file(1, OUT_PIPE);
	if(hWrite == INVALID_HANDLE_VALUE)
	{
		print_last_error();
		return -TIGLERR_DEV_OPEN_FAILED;
	}

	hRead = open_file(1, IN_PIPE);
	if(hRead == INVALID_HANDLE_VALUE)
	{
		print_last_error();
		return -TIGLERR_DEV_OPEN_FAILED;
	}

	return 0;
}

// Close device (libtifiles: slv_exit)
TIGLUSB_EXP int TIGLUSB_API TiglUsbClose(void)
{
	if((wBuf == NULL) || (rBuf == NULL))
	{
		free(wBuf); wBuf = NULL;
		free(rBuf); rBuf = NULL;
	}

	if(hWrite != INVALID_HANDLE_VALUE)
	{
		CloseHandle(hWrite);
		hWrite = INVALID_HANDLE_VALUE;
	}

	if(hRead != INVALID_HANDLE_VALUE)
	{
		CloseHandle(hRead);
		hRead = INVALID_HANDLE_VALUE;
	}

	if(hDev != INVALID_HANDLE_VALUE)
	{
		CloseHandle(hDev);
		hDev = INVALID_HANDLE_VALUE;
	}

	return 0;
}

// Reset both endpoints (libtifiles: slv_open)
TIGLUSB_EXP int TIGLUSB_API TiglUsbReset(void)
{
	nBytesRead = 0;
	nBytesWrite = 0;
	ioPending = FALSE;

	return (resetPipes(hDev) ? TIGLERR_RESET_FAILED : 0);
}

// send whole buffer at once
static int TiglUsbWriteBlockFull(UCHAR *data, UINT length)
{
	BOOL fSuccess;
	DWORD nBytesWritten;
	TIME clk;

	// start operation
	memset(&olp, 0, sizeof(OVERLAPPED));
	fSuccess=WriteFile(hWrite, data, length, &nBytesWritten, &olp);
	if( (fSuccess == FALSE) && (GetLastError() != ERROR_IO_PENDING) )
	{
		fprintf(stdout, "TiglUsb.dll: Write error.\n"); 
		return TIGLERR_WRITE_ERROR;
	}

	// and wait for completion (overlapped = asynchronous)
	toSTART(clk);
	do 
	{
		if(toELAPSED(clk, timeout)) 
		{ 
			fprintf(stdout, "TiglUsb.dll: Write timeout error.\n"); 
			return TIGLERR_WRITE_TIMEOUT; 
		}
	} 
	while(HasOverlappedIoCompleted(&olp) == FALSE);

	fSuccess = GetOverlappedResult(hWrite, &olp, &nBytesWritten, FALSE);
	if(fSuccess == FALSE)
	{
		fprintf(stdout, "TiglUsb.dll: Write error.\n"); 
		return TIGLERR_WRITE_ERROR;
	}
	if(nBytesWritten != length)
	{
		fprintf(stdout, "TiglUsb.dll: Write timeout error (%i).\n", timeout); 
		return TIGLERR_WRITE_TIMEOUT;
	}

	return 0;
}

// send whole buffer byte per byte
static int TiglUsbWriteBlockPart(UCHAR *data, UINT length)
{
	UINT i;

	for(i=0; i<length; i++) 
	{
		int ret = TiglUsbWriteBlockFull(&data[i], 1);
		if(ret) return ret;
	}

	return 0;
}


// Write one byte (libtifiles: slv_put)
TIGLUSB_EXP int TIGLUSB_API TiglUsbWrite(UCHAR data)
{
	// fill buffer (up to 32 bytes)
  	wBuf[nBytesWrite++] = data;

	// is buffer full? Then, send the whole buffer once
  	if (nBytesWrite == max_ps) 
	{
		int ret = TiglUsbWriteBlockFull(wBuf, nBytesWrite);
		nBytesWrite = 0;
        return ret;
	}

	return 0;
}

// Flush write buffer (libtifiles: slv_close)
TIGLUSB_EXP int TIGLUSB_API TiglUsbFlush(void)
{
	if (nBytesWrite > 0) 
	{
		int ret = TiglUsbWriteBlockPart(wBuf, nBytesWrite);
		nBytesWrite = 0;
		if(ret) return ret;
	}

	return 0;
}

// Check for data arrival (libtifiles: slv_check)
TIGLUSB_EXP int TIGLUSB_API TiglUsbCheck(int *status)
{
	BOOL fSuccess;

	*status = 0;

	if(nBytesRead > 0)
	{
		*status = !0;
		return 0;
	}

	if(ioPending == FALSE)
	{
		// start operation
		memset(&olp, 0, sizeof(OVERLAPPED));
		fSuccess = ReadFile(hRead, rBuf, max_ps, NULL, &olp);
		if( (fSuccess == FALSE) && (GetLastError() != ERROR_IO_PENDING) )
		{
			nBytesRead = 0;
			fprintf(stdout, "TiglUsb.dll: Read error.\n"); 
			return TIGLERR_READ_ERROR;
		}

		ioPending = TRUE;
	}
	else
	{
		// and monitor completion
		if(HasOverlappedIoCompleted(&olp) == FALSE)
			*status = 0;
		else
			*status = !0;
	}
	
	return 0;
}

// Read one byte (libtifiles: slv_get)
TIGLUSB_EXP int TIGLUSB_API TiglUsbRead(UCHAR *data)
{
	BOOL fSuccess;
	TIME clk;
	static UCHAR *rBufPtr;

  	/* 
		Flush write buffer by writing data byte per byte (more reliable)
	*/
	int ret = TiglUsbFlush();
  	if(ret) return ret;

	/* 
		This routine try to read up to 32 bytes (BULKUSB_MAX_TRANSFER_SIZE) and store them
		in a buffer for subsequent accesses 
	*/
	if(nBytesRead <= 0)
	{
retry:
		if(ioPending == FALSE)
		{
			// start operation
			memset(&olp, 0, sizeof(OVERLAPPED));
			fSuccess = ReadFile(hRead, rBuf, max_ps, &nBytesRead, &olp);
			if( (fSuccess == FALSE) && (GetLastError() != ERROR_IO_PENDING) )
			{
				nBytesRead = 0;
				fprintf(stdout, "TiglUsb.dll: Read error.\n"); 
				return TIGLERR_READ_ERROR;
			}

			// and wait for completion (overlapped = asynchronous)
			toSTART(clk);
			do 
			{
				if(toELAPSED(clk, timeout)) 
				{ 
					nBytesRead = 0;
					fprintf(stdout, "TiglUsb.dll: Read timeout error.\n"); 
					return TIGLERR_READ_TIMEOUT; 
				}
			} 
			//while(!GetOverlappedResult(hRead, &olp, &nBytesRead, FALSE));	//wait bytes are written
			while(HasOverlappedIoCompleted(&olp) == FALSE);
		}

		fSuccess = GetOverlappedResult(hRead, &olp, &nBytesRead, FALSE);
		ioPending = FALSE;

		if( (fSuccess == FALSE) && (GetLastError() != ERROR_IO_PENDING) )
		{
			nBytesRead = 0;
			fprintf(stdout, "TiglUsb.dll: Read error.\n"); 
			return TIGLERR_READ_ERROR;
		}

        // This is the quirk: sometimes returns with no data and no error; retry !
		if(nBytesRead == 0)
		{
			fprintf(stdout, "TiglUsb.dll: Read without data; retrying...\n");
			goto retry;
		}

		rBufPtr = rBuf;
	}
	
	*data = *rBufPtr++;
	nBytesRead--;

	return 0;
}

// Set timeout
TIGLUSB_EXP int TIGLUSB_API TiglUsbSetTimeout(int ts)
{
	timeout = ts;

	return 0;
}

// Get timeout
TIGLUSB_EXP int TIGLUSB_API TiglUsbGetTimeout(int *ts)
{
	if(ts != NULL)
		*ts = timeout;

	return timeout;
}
