// 
// TiglUsb driver for Windows 9x/Me, NT4/2000/XP
// Library (user mode) for talking with the driver (kernel mode)
//
// Copyright (c) 2001-2005 Romain Livin
// roms@lpg.ticalc.org
// http://lpg.ticalc.org/prj_usb
//
// July the 14th, 2005
//

/*++

Module Name:

    Device.c
	(this file is the same for both drivers (98DDK & XPDDK))

Abstract:

    Library (user mode) used for talking with the driver (kernel mode).
	This library contains a set of functions for:
	- opening/closing the driver
	- reading/writing with the driver
	- flushing buffers
	- setting/getting timeout
	- checking arrival of data (useful for emulators which need to monitor the linkport)
	- probing cables

Environment:

    user mode only

Notes:

  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
  KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
  PURPOSE.

  Copyright (c) 2001-2005 Romain Livin.  All Rights Reserved.


Revision History:

	03/07/02: created
	23/08/02: overlapped I/O fixes
	06/10/02: added a function for polling data arrival & improvements
	21/12/02: added a function for retrieving library version
    11/04/04: improved to workaround the quirk (returns neither error nor data)
	21/04/04: improved to use heuristic-buffered writes (send is faster but still reliable)
	01/05/05: architecture changed (use handles) for supporting multiple cables.
			  added extended functions for sending/receiving whole packets
	17/07/05: rewrite of this API starting at the old one due to low data-rate 
		and bad resumability of transfers. Don't ask me why !
	10/09/05: don't use the same olp structure for R & W operations
--*/

/* 
   Some important remarks from <http://lpg.ticalc.org/prj_usb/SilverLink_Paper.txt>
   
   This link cable use Bulk mode with packets. The max size of a packet is 
   32/64 bytes (32 on SilverLink, 64 for hand-helds). 
   
   This is transparent for the user because the libusb manages all these 
   things for us. Nevertheless, this fact has some consequences:

   - it is better (for USB & OS performances) to read/write a set of bytes 
   rather than byte per byte.

   - for reading, we have to read up to 32 bytes at a time (even if we need 
   only 1 byte) and to store them in a buffer for subsequent accesses. 
   In fact, if we try and get byte per byte, it will not work.

   - [ for writing, we don't store bytes in a buffer. It seems better to send
   data byte per byte (latency ?!). But, this make data-rate significantly 
   decrease (1KB/s instead of 5KB/s).
   Another way is to use partially buffered write operations: send consecutive
   blocks as a whole but partial block byte per byte. This is the best 
   compromise and it works fine ! ] => seems to be un-needed for now ?!

   - another particular effect (quirk): sometimes (usually when calc need to 
   reply and takes a while), a read call can returns with neither data 
   nor timeout. Simply retry a read call and it works fine. The best example 
   is to get IDLIST.
*/

#include <stdio.h>
#include <stdlib.h>

#include "stdafx.h"
#include "TiglUsb.h"

//
// TiglUsb.cpp: DLL routines
//

typedef DWORD                   TIME;
# define  toSTART(ref)          { (ref)=GetTickCount(); }
# define  toELAPSED(ref, max)   ( (int)(GetTickCount()-(ref)) > (100*max) )

static TiglUsbHandle* handles[USB_MAX] = { 0 };

#define number	(handles[h]->number)
#define max_ps	(handles[h]->max_ps)
#define timeout (handles[h]->timeout)

#define hRead   (handles[h]->hRead)
#define hWrite  (handles[h]->hWrite)

#define nBytesRead	(handles[h]->nBytesRead)
#define rBuf	(handles[h]->rBuf)
#define rBufPtr	(handles[h]->rBufPtr)

#define olpr	(handles[h]->olpr)
#define olpw	(handles[h]->olpw)

#define hDev	(handles[h]->hDev)
#define pid		(handles[h]->pid)

static BOOL ioPending = FALSE;

static int handle_new(void)
{
	int i;

	for(i = 0; i < USB_MAX; i++)
		if(handles[i] == NULL)
		{
			handles[i] = (TiglUsbHandle *)calloc(1, sizeof(TiglUsbHandle));
			return i;
		}

	return -1;
}

static void handle_del(int i)
{
	free(handles[i]);
	handles[i] = NULL;
}

static void print_last_error(void)
{
	LPVOID lpMsgBuf;
	
	FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
					FORMAT_MESSAGE_FROM_SYSTEM |
					FORMAT_MESSAGE_IGNORE_INSERTS,
					NULL, GetLastError(),
					MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
					(LPTSTR) &lpMsgBuf,    0,    NULL );
	fprintf(stdout, "TiglUsb.dll error: %i -> %s\n", GetLastError(), lpMsgBuf);
}

// Prove available devices
TIGLUSB_EXP INT TIGLUSB_API TiglUsbProbe2(PUINT* list)
{
	return (int)list_dev(list);
}

// Open device
TIGLUSB_EXP int TIGLUSB_API TiglUsbOpen2(int id)
{
	int h = 0;

	// Allocate an handle
	h = handle_new();
	if(h == -1)
		return TIGLERR2_HANDLE_MAX;
	timeout = 15;

	// Get maximum packet size & Product ID
	hDev = open_dev(number = id);
	if(hDev == INVALID_HANDLE_VALUE)
	{
		print_last_error();
		return TIGLERR2_DEV_OPEN_FAILED;
	}

	if(get_max_ps(hDev, &max_ps, &pid) == -1)
		return TIGLERR2_DEV_OPEN_FAILED;
	
	// Open the USB device: 2 named pipes (endpoints)
	hWrite = open_file(id, OUT_PIPE);
	if(hWrite == INVALID_HANDLE_VALUE)
	{
		print_last_error();
		return TIGLERR2_DEV_OPEN_FAILED;
	}

	hRead = open_file(id, IN_PIPE);
	if(hRead == INVALID_HANDLE_VALUE)
	{
		print_last_error();
		return TIGLERR2_DEV_OPEN_FAILED;
	}

	return h;
}

// Close device
TIGLUSB_EXP int TIGLUSB_API TiglUsbClose2(int h)
{
	// Close handles
	if(hRead != INVALID_HANDLE_VALUE)
	{
		CloseHandle(hRead);
		hRead = INVALID_HANDLE_VALUE;
	}

	if(hWrite != INVALID_HANDLE_VALUE)
	{
		CloseHandle(hWrite);
		hWrite = INVALID_HANDLE_VALUE;
	}

	// Release library
	if(hDev != INVALID_HANDLE_VALUE)
	{
		CloseHandle(hDev);
		hDev = INVALID_HANDLE_VALUE;
	}

	// Release handle
	handle_del(h);

	return 0;
}

// Reset both endpoints
TIGLUSB_EXP int TIGLUSB_API TiglUsbReset2(int h)
{
	BOOL fSuccess;
	int ret;
	//int to;
	//UCHAR buf[1024];

	fSuccess = CancelIo(hRead);
	if(fSuccess == FALSE)
	{
		fprintf(stdout, "TiglUsb.dll: CancelIo error.\n"); 
		return TIGLERR2_RESET_FAILED;
	}

	fSuccess = CancelIo(hWrite);
	if(fSuccess == FALSE)
	{
		fprintf(stdout, "TiglUsb.dll: CancelIo error.\n"); 
		return TIGLERR2_RESET_FAILED;
	}

	nBytesRead = 0;
	rBufPtr = rBuf;
	ioPending = FALSE;

	ret = (resetPipes(hDev) ? TIGLERR2_RESET_FAILED : 0);
/*
	to = timeout;
	timeout = 1;
	TiglUsbReads2(h, buf, sizeof(buf));
	timeout = to;
*/
	return ret;
}

// send whole buffer at once
static int TiglUsbWriteBlockFull(INT h, UCHAR *data, UINT length)
{
	BOOL fSuccess;
	DWORD nBytesWritten;
	TIME clk;

	// start operation
	memset(&olpw, 0, sizeof(OVERLAPPED));
	fSuccess = WriteFile(hWrite, data, length, &nBytesWritten, &olpw);
	if( (fSuccess == FALSE) && (GetLastError() != ERROR_IO_PENDING) )
	{
		fprintf(stdout, "TiglUsb.dll: Write error.\n"); 
		return TIGLERR2_WRITE_ERROR;
	}

	// and wait for completion (overlapped = asynchronous)
	toSTART(clk);
	do 
	{
		if(toELAPSED(clk, timeout)) 
		{ 
			fprintf(stdout, "TiglUsb.dll: Write timeout error (%i).\n", timeout); 
			return TIGLERR2_WRITE_TIMEOUT; 
		}
	} 
	while(HasOverlappedIoCompleted(&olpw) == FALSE);

	fSuccess = GetOverlappedResult(hWrite, &olpw, &nBytesWritten, FALSE);
	if(fSuccess == FALSE)
	{
		fprintf(stdout, "TiglUsb.dll: Write error.\n"); 
		return TIGLERR2_WRITE_ERROR;
	}
	if(nBytesWritten != length)
	{
		fprintf(stdout, "TiglUsb.dll: Write timeout error (%i).\n", timeout); 
		return TIGLERR2_WRITE_TIMEOUT;
	}

	return 0;
}

// Write one byte (beware: slower than TiglUsbWrites2)
TIGLUSB_EXP int TIGLUSB_API TiglUsbWrite2(INT h, UCHAR data)
{
	return TiglUsbWriteBlockFull(h, &data, 1);
}

// Write buffer
TIGLUSB_EXP int TIGLUSB_API TiglUsbWrites2(INT h, UCHAR *data, INT len)
{
	return TiglUsbWriteBlockFull(h, data, len);
}

// Does nothing
TIGLUSB_EXP int TIGLUSB_API TiglUsbFlush2(INT h)
{
	return 0;
}

// Check for data arrival
TIGLUSB_EXP int TIGLUSB_API TiglUsbCheck2(INT h, int *status)
{
	BOOL fSuccess;

	*status = 0;

	if(nBytesRead > 0)
	{
		*status = !0;
		return 0;
	}

	if(ioPending == FALSE)
	{
		// start operation
		memset(&olpr, 0, sizeof(OVERLAPPED));
		fSuccess = ReadFile(hRead, rBuf, max_ps, NULL, &olpr);
		if( (fSuccess == FALSE) && (GetLastError() != ERROR_IO_PENDING) )
		{
			nBytesRead = 0;
			fprintf(stdout, "TiglUsb.dll: Read error.\n"); 
			return TIGLERR2_READ_ERROR;
		}

		ioPending = TRUE;
	}
	else
	{
		// and monitor completion
		if(HasOverlappedIoCompleted(&olpr) == FALSE)
			*status = 0;
		else
			*status = !0;
	}

	return 0;
}

// Read one byte (buffered)
static int TiglUsbRead2(INT h, UCHAR *data)
{
	BOOL fSuccess;
	TIME clk;
	//static UCHAR *rBufPtr2;

	/* 
		This routine try to read up to 32/64 bytes and store them
		in a buffer for subsequent accesses (caching)
	*/
	if(nBytesRead <= 0)
	{
retry:
		if(ioPending == FALSE)
		{
			// start operation
			memset(&olpr, 0, sizeof(OVERLAPPED));
			fSuccess = ReadFile(hRead, rBuf, max_ps, &nBytesRead, &olpr);
			if( (fSuccess == FALSE) && (GetLastError() != ERROR_IO_PENDING) )
			{
				nBytesRead = 0;
				fprintf(stdout, "TiglUsb.dll: Read error.\n"); 
				return TIGLERR2_READ_ERROR;
			}

			// and wait for completion (overlapped = asynchronous)
			toSTART(clk);
			do 
			{
				if(toELAPSED(clk, timeout)) 
				{ 
					nBytesRead = 0;
					fprintf(stdout, "TiglUsb.dll: Read timeout error (%i).\n", timeout); 
					return TIGLERR2_READ_TIMEOUT; 
				}
			} 
			while(HasOverlappedIoCompleted(&olpr) == FALSE);
		}

		fSuccess = GetOverlappedResult(hRead, &olpr, &nBytesRead, FALSE);
		ioPending = FALSE;

		if( (fSuccess == FALSE) && (GetLastError() != ERROR_IO_PENDING) )
		{
			nBytesRead = 0;
			fprintf(stdout, "TiglUsb.dll: Read error.\n"); 
			return TIGLERR2_READ_ERROR;
		}

        // This is the quirk: sometimes returns with no data and no error; retry !
		if(nBytesRead == 0)
		{
			fprintf(stdout, "TiglUsb.dll: Read without data; retrying...\n");
			goto retry;
		}

		rBufPtr = rBuf;
	}
	
	*data = *rBufPtr++;
	nBytesRead--;

	return 0;
}

// Read buffer
TIGLUSB_EXP int TIGLUSB_API TiglUsbReads2(INT h, UCHAR *data, INT len)
{
	int i;

	// we can't do that in any other way because TiglUsbRead can returns 
	// 1, 2, ..., len bytes.
	for(i = 0; i < len; i++)
	{
		int ret = TiglUsbRead2(h, data+i);
		if (ret) return ret;		
	}

	return 0;
}

// Set timeout
TIGLUSB_EXP int TIGLUSB_API TiglUsbSetTimeout2(INT h, int ts)
{
	int old = timeout;

	timeout = ts;
	//fprintf(stdout, "TiglUsb.dll: timeout set to %i\n", ts);

	return old;
}

// Get timeout
TIGLUSB_EXP int TIGLUSB_API TiglUsbGetTimeout2(INT h, int *ts)
{
	if(ts != NULL)
		*ts = timeout;

	return timeout;
}
